'''
数据清洗过滤类转换算法
'''
import pandas as pd
from ezetl.utils.common_utils import trans_rule_value, parse_to_list


def empty_to_null(source_data={}, rule_dict={}, context={}):
    '''
    空字符串转null
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "处理字段列表",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields', ''))
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
               if fields == []:
                   fields = source_data[i].keys()
               for k in fields:
                   if k in source_data[i] and source_data[i][k] == '':
                       source_data[i][k] = None
    elif isinstance(source_data, dict):
        if fields == []:
            fields = source_data.keys()
        for k in fields:
            if k in source_data and source_data[k] == '':
                source_data[k] = None
    else:
        # dataframe处理
        if fields == []:
            fields = source_data.columns
        for k in fields:
            source_data[k] = source_data[k].apply(lambda x: None if x == '' else x)
    return True, source_data


def clean_empty(source_data={}, rule_dict={}, context={}):
    '''
    清除空字符串和null
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "处理字段列表",
        "value": "fields",
        "form_type": "select_fields",
        "required": true,
        "default": "",
        "tips": ""
    }]
    :return:
    '''
    fields = parse_to_list(rule_dict.get('fields', ''))
    if isinstance(source_data, list):
        for i in range(len(source_data)):
            if isinstance(source_data[i], dict):
               if fields == []:
                   fields = source_data[i].keys()
               for k in fields:
                   if k in source_data[i] and (source_data[i][k] == '' or source_data[i][k] is None):
                       source_data[i].pop(k)
    elif isinstance(source_data, dict):
        if fields == []:
            fields = source_data.keys()
        for k in fields:
            if k in source_data and (source_data[k] == '' or source_data[k] is None):
                source_data.pop(k)
    else:
        return False, '本算法对dataframe类型无效！'
    return True, source_data


def filter_by_rules(source_data={}, rule_dict={}, context={}):
    '''
    根据条件过滤
    :param source_data: 原始数据
    :param rule_dict: 配置信息
    [{
        "name": "过滤条件",
        "value": "filter_rules",
        "form_type": "filter_rules",
        "type": "list",
        "must": true,
        "default": [],
        "tips": ""
    }]
    :return:
    '''
    # dataframe处理
    filter_rules = rule_dict.get('filter_rules', [])
    if filter_rules == []:
        return False, '筛选条件不能为空'
    if isinstance(source_data, list):
        df = pd.DataFrame(source_data)
    elif isinstance(source_data, dict):
        df = pd.DataFrame([source_data])
    else:
        df = source_data
    for i in filter_rules:
        print(i)
        field = i.get('field')
        rule = i.get('rule')
        value = i.get('value')
        value = trans_rule_value(value)
        if field and value:
            if rule == 'equal':
                df = df[df[field] == value]
            elif rule == 'f_equal':
                df = df[df[field] != value]
            elif rule == 'gt':
                df = df[df[field] > value]
            elif rule == 'gte':
                df = df[df[field] >= value]
            elif rule == 'lt':
                df = df[df[field] < value]
            elif rule == 'lte':
                df = df[df[field] <= value]
    return True, df



from ezetl.utils import get_reader

reader_info = {
    'source': {
        "type": "file",
        "conn_conf": {
            "path":  "../test_files/btc_history.csv",
        },
    },
    'model': {
        "name": "test",
        "type": "file_table",
        "model_conf": {},
        "fields": []
    },
    'extract_info': {
        'batch_size': 10,
        'extract_rules': {'equal[]': 'setosa'}
    }
}

flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
print(reader.get_res_fields())
flag, read_data = reader.read_page()
print(flag, read_data)
for flag, i in reader.read_batch():
    print(flag, i)



from ezetl.libs.es import EsClient
from ezetl.data_models import DataModel
from ezetl.utils.common_utils import trans_rule_value, parse_json
from ezetl.utils.es_utils import get_index_mapping
from ezetl.utils.es_query_tool import EsQueryTool


class EsIndexModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        model_conf = self._model.get('model_conf', {})
        self.index_name = model_conf.get('name', '')
        self.auth_types = model_conf.get('auth_type', '').split(',')
        conn_conf = self._source['conn_conf']
        url = conn_conf.get('url')
        hosts = url.split(',')
        es_hosts = []
        for h in hosts:
            l = h.split(':')
            es_hosts.append({'host': l[0], 'port': l[1]})
        es_conf = {
            'hosts': es_hosts
        }
        auth_type = conn_conf.get('auth_type', 1)
        if auth_type == 2:
            # 用户名密码验证
            username = conn_conf.get('username')
            password = conn_conf.get('password')
            es_conf['http_auth'] = (username, password)
        self.es_client = EsClient(**es_conf)

    def connect(self):
        '''
        连通性测试
        '''
        try:
            mapping = self.es_client.get_mapping(self.index_name)
            if self.index_name != '':
                if self.index_name in mapping:
                    return True, '连接成功'
                else:
                    return False, '连接失败'
            else:
                return True, '连接成功'
        except Exception as e:
            return False, str(e)[:100]

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        try:
            mapping = get_index_mapping(self.index_name, self.es_client)
            field_dict = mapping[self.index_name]['mappings']['properties']
            res_fields = []
            for field_value in field_dict:
                field = field_dict[field_value]
                field_type = field.get('type')
                dic = {
                    'field_name': field_value,
                    'field_value': field_value,
                    'ext_params': {
                        'type': field_type,
                        'length': 0,
                        'is_primary_key': 0,
                        'nullable': 1,
                        'default': ''
                    }
                }
                res_fields.append(dic)
            return res_fields
        except Exception as e:
            print(e)
            return []

    def check_field(self, field_info, res_fields):
        '''
        检察字段是否存在且一致
        '''
        for field in res_fields:
            if field['field_value'] == field_info['field_value'] and field['type'] == field_info['type']:
                print(field_info, field)
                return True
        return False

    def set_field(self, field):
        '''
        设置字段
        '''
        if 'edit_fields' not in self.auth_types:
            return False, '无操作字段权限'
        mapping = {}
        mapping['properties'] = {}
        mapping['properties'][field['field_value']] = {}
        mapping['properties'][field['field_value']]['type'] = field['type']
        # if field['type'] == 'date':
        #     mapping['properties'][field['value']]['format'] = "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis"
        try:
            self.es_client.create_mapping(self.index_name, mapping)
            return True, '操作成功'
        except Exception as e:
            print(e)
            return False, str(e)[:100]

    def create(self, field_arr):
        '''
        创建索引
        '''
        if 'create' not in self.auth_types:
            return False, '无创建权限'
        mapping = {}
        mapping['properties'] = {}
        for field in field_arr:
            mapping['properties'][field['field_name']] = {}
            mapping['properties'][field['field_name']]['type'] = field['type']
            # if field['type'] == 'date':
            #     mapping['properties'][field['name']]['format'] = "yyyy-MM-dd HH:mm:ss||yyyy-MM-dd||epoch_millis"
        try:
            self.es_client.create_mapping(self.index_name, mapping)
            self.es_client.put_settings(self.index_name, {'refresh_interval': "1s", "max_result_window": "100000"})
            return True, '创建成功'
        except Exception as e:
            return False, str(e)

    def delete(self):
        '''
        删除索引
        '''
        if 'delete' not in self.auth_types:
            return False, '无删除权限'
        try:
            self.es_client.delete_index(self.index_name)
            return True, '删除成功'
        except Exception as e:
            return False, str(e)

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return [{
            'name': '关键词',
            'value': 'search_key',
            "default": f""
        }, {
            'name': 'sql',
            'value': 'sql',
            "default": f"select * from {self.index_name}"
        }]

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal'
        }, {
            'name': '不等于',
            'value': 'f_equal'
        }, {
            'name': '包含',
            'value': 'contain'
        }, {
            'name': '不包含',
            'value': 'f_contain'
        }, {
            'name': '大于',
            'value': 'gt'
        }, {
            'name': '大于等于',
            'value': 'gte'
        }, {
            'name': '小于',
            'value': 'lt'
        }, {
            'name': '小于等于',
            'value': 'lte'
        }, {
            'name': '从大到小排序',
            'value': 'sort_desc'
        }, {
            'name': '从小到大排序',
            'value': 'sort_asc'
        }, {
            'name': '统计数量',
            'value': 'jl_terms'
        }, {
            'name': '统计数量页数',
            'value': 'jl_page_size'
        }, {
            'name': '聚合统计',
            'value': 'jl_stats'
        }, {
            'name': '统计总和',
            'value': 'jl_sum'
        }, {
            'name': '统计平均值',
            'value': 'jl_avg'
        }, {
            'name': '统计最小值',
            'value': 'jl_min'
        }, {
            'name': '统计最大值',
            'value': 'jl_max'
        }, {
            'name': '返回字段列表',
            'value': 'return_fields'
        }, {
            'name': '唯一号列表筛选',
            'value': 'doc_ids'
        }, {
            'name': '返回内容',
            'value': 'content_tag'
        }
        ]
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        api_form = {}
        jl_tag = []
        for i in self.extract_rules:
            print(i)
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if rule in ['equal', 'eq']:
                api_form[f'equal[{field}]'] = value
            elif rule in ['f_equal', 'neq']:
                api_form[f'f_equal[{field}]'] = value
            elif rule in ['gt', 'gte', 'lt', 'lte', 'contain', 'f_contain', 'jl_page_size']:
                api_form[f'{rule}[{field}]'] = value
            elif rule in ['return_fields', 'doc_ids', 'search_key']:
                api_form[rule] = value
            elif rule == 'sort':
                if field:
                    api_form[f'sort[{field}]'] = value
                else:
                    api_form[f'sort'] = value
            elif rule == 'sort_asc':
                api_form[f'sort[{field}]'] = 'ASC'
            elif rule == 'sort_desc':
                api_form[f'sort[{field}]'] = 'DESC'
            elif rule in ['jl_terms', 'jl_sum', 'jl_avg', 'jl_max', 'jl_min', 'jl_stats']:
                api_form[f'jl_type[{field}]'] = rule.split('_')[-1]
                jl_tag.append(field)
            elif rule == 'content_tag':
                api_form['content_tag'] = value
            elif rule == 'search_text':
                if field == 'sql':
                    api_form['search_key'] = f"sql:{value}"
                elif field == 'search_key':
                    api_form['search_key'] = value
            elif rule == 'jl_tag':
                jl_tag.extend(value.split(','))
            elif rule == 'jl_type':
                api_form[f'jl_type[{field}]'] = value
                jl_tag.append(field)
        if jl_tag != []:
            api_form['jl_tag'] = ','.join(list(set(jl_tag)))
        return api_form

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        api_form = self.gen_extract_rules()
        api_form['index_name'] = self.index_name
        print(api_form)
        api_form['page'] = page
        api_form['pagesize'] = pagesize
        es_tools = EsQueryTool(api_form)
        res_data = es_tools.query(es=self.es_client)
        return True, res_data

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        api_form = self.gen_extract_rules()
        api_form['index_name'] = self.index_name
        print(api_form)
        # scroll 获取数据
        es_tools = EsQueryTool(api_form)
        query_body = es_tools.query_body
        if query_body['query'] == {}:
            query_body['query'] = {'match_all': {}}
        query_body['size'] = self._extract_info.get('batch_size', 1000)
        result = self.es_client.scroll_search(self.index_name, query_body)
        sid = result['_scroll_id']
        scroll_size = result['hits']['total']['value']
        print(scroll_size)
        while scroll_size > 0:
            res_data = es_tools.gen_result(result)
            yield True, res_data
            result = self.es_client.scroll(sid)
            sid = result['_scroll_id']
            if res_data['data']['records'] == []:
                break

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert', 'update', 'upsert']:
            return False, '写入类型参数错误'
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            if self.load_type == 'insert':
                self.es_client.add_data_bulk(self.index_name, records)
            elif self.load_type == 'update':
                self.es_client.update_data_bulk(self.index_name, records)
            elif self.load_type == 'upsert':
                self.es_client.update_data_bulk(self.index_name, records, upsert=True)
        except Exception as e:
            return False, f'{str(e)[:-100]}'
        return True, records



from ezetl.data_models.base_db_table import BaseDBTableModel
from sqlalchemy import Column
from sqlalchemy.dialects.mysql import LONGTEXT


def gen_longtext_column(i):
    if 'nullable' in i.keys():
        i['nullable'] = i['nullable'] == 1
    else:
        i['nullable'] = True
    obj = Column(LONGTEXT, nullable=i['nullable'])
    return obj


class MysqlTableModel(BaseDBTableModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        self.column_gen_map = {
            'LONGTEXT': gen_longtext_column
        }



from setuptools import setup, find_packages

version = '1.0.2'

setup(
    name='ez-etl',
    version=version,
    description="ez-etl is an open-source Extract, Transform, load (ETL) library written in Python. Just configure a dict to read data from various data sources. Use code or built-in conversion algorithms to transform the data into the target data format， and write it to the target data source. ",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
        ],
    },
    author='Xu Wei',
    author_email='1013104194@qq.com',
    license='Apache License 2.0',
    install_requires=[
        "requests==2.25.1",
        "sshtunnel==0.4.0",
        "python-dateutil==2.8.2",
        "pandas>=1.0.0",
        # "xorbits==0.3.2",
        # "pyarrow>=5.0.0",
        "openpyxl==3.0.9",
        "xlrd==1.2.0",
        "minio==7.1.1",
        "sqlalchemy>=1.2.0",
        "pymysql==1.0.2",
        "elasticsearch>=7.15.2",
        "redis==4.0.2",
        "clickhouse-driver==0.2.3",
        "clickhouse-sqlalchemy== 0.2.0",
        "influxdb==5.3.1",
        "py2neo==2020.0.0",
        "kafka-python==2.0.2",
        "mongoengine==0.24.2",
        "mysql-replication==0.27"
    ],
    keywords="etl, extract, transform, load, excel, csv, mysql, kafka, elasticsearch, neo4j, influxdb, mongodb, clickhouse",
)


from ezetl.data_models.base_db_table import BaseDBTableModel
from clickhouse_sqlalchemy import engines
from sqlalchemy import Column, TEXT, String


def gen_fixstring_column(i):
    if 'nullable' in i.keys():
        i['nullable'] = i['nullable'] == 1
    else:
        i['nullable'] = True
    if 'primary_key' in i.keys():
        i['primary_key'] = i['primary_key'] == 1
    else:
        i['primary_key'] = True
    obj = Column(String(i['length']), nullable=i.get('nullable', True), primary_key=i.get('primary_key', False))
    return obj


def gen_string_column(i):
    if 'nullable' in i.keys():
        i['nullable'] = i['nullable'] == 1
    else:
        i['nullable'] = True
    obj = Column(TEXT, nullable=i['nullable'])
    return obj


class CkTableModel(BaseDBTableModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        self.column_gen_map = {
            'FixedString': gen_fixstring_column,
            'String': gen_string_column
        }
        model_conf = self._model.get('model_conf', {})
        engine = model_conf.get('engine', 'MergeTree')
        order_by = model_conf.get('order_by', ['id'])
        if engine and engine == 'ReplacingMergeTree':
            self.table_args = (
                engines.ReplacingMergeTree(order_by=order_by),
            )
        else:
            self.table_args = (
                engines.MergeTree(order_by=order_by),
            )

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert']:
            return False, f'写入类型参数错误,不支持类型{self.load_type}'
        if self.table is None:
            return False, '表不存在'
        columns = self.table.columns
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            insert_records = records
            # 创建 insert 对象
            ins = self.table.insert()
            conn = self.db_engine.connect()
            tmp_list = [{k: v for k, v in c.items() if k in columns} for c in insert_records]
            conn.execute(ins, tmp_list)
        except Exception as e:
            return False, f'{str(e)[:100]}'
        return True, res_data


from ezetl.utils import get_reader

reader_info = {
    'source': {
        "name": "test",
        "type": "kafka",
        "conn_conf": {
            "bootstrap_servers": "127.0.0.1:9092",
        },
        "ext_params": {}
    },
    'model': {
        "name": "test",
        "type": "kafka_topic",
        "model_conf": {
            "name": "btc_history",
        },
        "ext_params": {
            "group_id": "111"
        },
        "fields": []
    },
    'extract_info': {
        'batch_size': 1,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
for i in reader.read_batch():
    print(i)


from ezetl.utils import get_reader, get_writer

reader_info = {
    'source': {
        "name": "test",
        "type": "mongodb",
        "conn_conf": {
            "database_name": "datacenter",
            "host": "127.0.0.1",
            "port": "27017",
            "username": "root",
            "password": "123456",
        },
        "ext_params": {}
    },
    'model': {
        "name": "test",
        "type": "mongodb_collection",
        "model_conf": {
            "name": "btc_history",
        },
        "ext_params": {},
        "fields": []
    },
    'extract_info': {
        'batch_size': 100,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
print(reader.get_res_fields())
read_data = reader.read_page()
print(read_data)
for i in reader.read_batch():
    print(i)




from ezetl.transform_algs import filter_algs, count_algs, map_algs, content_algs
from ezetl.transform_algs.content_algs import df_to_data

# 算法字典
transform_alg_dict = {
    # 内容提取类算法
    'code_transform': content_algs.code_transform,
    'gen_records_list': content_algs.gen_records_list,
    'data_to_df': content_algs.data_to_df,
    'df_to_data': content_algs.df_to_data,
    'gen_es_aggs_buckets': content_algs.gen_es_aggs_buckets,
    'gen_es_aggs_value': content_algs.gen_es_aggs_value,
    'gen_contents_first': content_algs.gen_contents_first,
    'gen_contents_total': content_algs.gen_contents_total,
    # 映射类算法
    'map_field_names': map_algs.map_field_names,
    'map_values': map_algs.map_values,
    'trans_time_format': map_algs.trans_time_format,
    'trans_field_type': map_algs.trans_field_type,
    'gen_only_id': map_algs.gen_only_id,
    'add_field': map_algs.add_field,
    # 数据清洗类算法
    'clean_empty': filter_algs.clean_empty,
    'empty_to_null': filter_algs.empty_to_null,
    # 统计聚合类算法
    'group_agg_count': count_algs.group_agg_count
}



from ezetl.data_models import DataModel
from ezetl.utils.db_utils import get_database_engine, get_database_model
from ezetl.utils.common_utils import trans_rule_value, gen_json_response, parse_json
from sqlalchemy import not_
from sqlalchemy.sql.schema import Table
from sqlalchemy import Column, String, Integer, Text, SmallInteger, DateTime, TIMESTAMP, Float
from sqlalchemy.ext.declarative import declarative_base


class BaseDBTableModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        self.db_type = self._source.get('type')
        self.conn_conf = self._source['conn_conf']
        self.conn_conf['type'] = self.db_type
        model_conf = self._model['model_conf']
        self.table_name = model_conf.get('name')
        self.auth_types = model_conf.get('auth_type', '').split(',')
        self.column_gen_map = {}
        Base = declarative_base()
        self.get_table_model()

        class BaseModel(Base):
            '''
            数据库orm模型基类
            '''
            __abstract__ = True
            id = Column(String(36), primary_key=True, nullable=False, comment='ID')

        self.BaseModel = BaseModel
        self.table_args = None

    def get_table_model(self):
        '''
        获取表模型
        :return:
        '''
        self.db_engine = get_database_engine(self.conn_conf)
        self.session, self.db_model = get_database_model(self.table_name, self.db_engine, db_type=self.db_type)
        if isinstance(self.db_model, Table):
            self.table = self.db_model
        elif self.db_model:
            self.table = self.db_model.__table__
        else:
            self.table = None

    def connect(self):
        '''
        连通性测试
        '''
        try:
            if self.table is not None:
                return True, '连接成功'
            else:
                return False, '连接失败'
        except Exception as e:
            return False, str(e)[:100]

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        res_fields = []
        try:
            for column in self.table.columns:
                try:
                    length = column.type.length if column.type.length else 0
                except:
                    length = 0
                column_type = str(column.type)
                for t in ['VARCHAR', 'FixedString']:
                    if t in column_type:
                        column_type = t
                        break
                dic = {
                    'field_name': column.comment if column.comment else column.name,
                    'field_value': column.name,
                    'ext_params': {
                        'type': column_type,
                        'length': length,
                        'is_primary_key': column.primary_key,
                        'nullable': column.nullable,
                        'default': column.default
                    }
                }
                res_fields.append(dic)
        except Exception as e:
            print(e)
        return res_fields

    def genColumn(self, i):
        '''
        根据字典获取列参数，组成orm模型字段
        :param i:
        :return:
        '''
        if 'nullable' in i.keys():
            i['nullable'] = i['nullable'] == 1
        else:
            i['nullable'] = True
        if 'primary_key' in i:
            i['is_primary_key'] = i['primary_key'] == 1
        else:
            i['is_primary_key'] = False
        if 'length' not in i:
            i['length'] = '0'
        if 'type' not in i:
            obj = Column(Text, nullable=True)
        elif i['type'] in ['varchar', 'String']:
            obj = Column(String(i['length']), primary_key=i['is_primary_key'], nullable=i['nullable'])
        elif i['type'] in ['text', 'Text']:
            obj = Column(Text, nullable=i['nullable'])
        elif i['type'] in ['int', 'Int', 'Integer']:
            obj = Column(Integer, primary_key=i['is_primary_key'], nullable=i['nullable'])
        elif i['type'] in ['float', 'Float', 'FLOAT']:
            obj = Column(Float, nullable=i['nullable'])
        elif i['type'] == ['smallint', 'SmallInteger']:
            obj = Column(SmallInteger, nullable=i['nullable'])
        elif i['type'] in ['datetime', 'DateTime']:
            obj = Column(DateTime, nullable=i['nullable'])
        elif i['type'] in ['timestamp', 'TimeStamp', 'TIMESTAMP']:
            obj = Column(TIMESTAMP, nullable=i['nullable'])
        else:
            obj = Column(Text, nullable=True)
        if 'field_name' in i:
            obj.comment = i['field_name']
        return obj

    def create(self, field_arr=[]):
        '''
        创建表
        '''
        if 'create' not in self.auth_types:
            return False, '无创建权限'
        table_args = self.table_args
        try:
            table_name = self.table_name

            class Model(self.BaseModel):
                __tablename__ = self.table_name
                if table_args is not None:
                    __table_args__ = table_args

            for i in field_arr:
                if i['field_value'] == 'id':
                    continue
                if i['type'] in self.column_gen_map:
                    column = self.column_gen_map[i['type']](i)
                else:
                    column = self.genColumn(i)
                setattr(Model, i['field_value'], column)

            from sqlalchemy.schema import CreateTable
            print(CreateTable(Model.__table__))
            if table_name not in self.db_engine.table_names():
                Model.__table__.create(self.db_engine)
            self.db_engine.dispose()
            # 重新获取模型
            self.get_table_model()
            return True, '创建成功'
        except Exception as e:
            return False, str(e)

    def delete(self):
        '''
        删除表
        '''
        if 'delete' not in self.auth_types:
            return False, '无删除权限'
        try:
            self.table.drop(self.db_engine)
            self.db_engine.dispose()
            return True, '删除表成功'
        except Exception as e:
            return False, str(e)

    def check_field(self, field_info, res_fields):
        '''
        检察字段是否存在且一致
        '''
        for field in res_fields:
            if field['field_value'] == field_info['field_value'] and field['type'] == field_info['type'] and field['length'] == field_info['length'] and field['nullable'] == field_info['nullable']:
                print(field_info, field)
                return True
        return False

    def set_field(self, field):
        '''
        设置字段
        '''
        if 'edit_fields' not in self.auth_types:
            return False, '无操作字段权限'
        try:
            set_type = 'ADD'
            for column in self.table.columns:
                if column.name == field['field_name']:
                    set_type = 'MODIFY'
                    break
            if field['type'] in ['varchar', 'VARCHAR', 'FixedString']:
                field['type'] = f"{field['type']}({field['length']})"
            c_sql = f"""ALTER TABLE `{self.table_name}` {set_type} COLUMN `{field['field_name']}` {field['type']} COMMENT '{field.get('comment', '')}';"""
            print(c_sql)
            self.db_engine.execute(c_sql)
            return True, '操作成功'
        except Exception as e:
            return False, str(e)

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal',
          }, {
            'name': '不等于',
            'value': 'f_equal'
          }, {
            'name': '包含',
            'value': 'contain'
          }, {
            'name': '不包含',
            'value': 'f_contain'
          }, {
            'name': '大于',
            'value': 'gt'
          }, {
            'name': '大于等于',
            'value': 'gte'
          }, {
            'name': '小于',
            'value': 'lt'
          }, {
            'name': '小于等于',
            'value': 'lte'
          }, {
            'name': '从大到小排序',
            'value': 'sort_desc'
          }, {
            'name': '从小到大排序',
            'value': 'sort_asc'
          }
        ]
        return rules

    def gen_extract_rules(self, model):
        '''
        解析筛选规则
        :return:
        '''
        query = self.session.query(model)
        if self.table is None:
            return False, '表不存在'

        columns = self.table.columns
        for i in self.extract_rules:
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if field and field not in ['sql', 'search_text']:
                column = columns.get(field)
                if column is None:
                    return False, f'未找到字段：{field}'
                if value:
                    if rule in ['eq', 'equal']:
                        query = query.filter(column == value)
                    elif rule == ['neq', 'f_equal']:
                        query = query.filter(column != value)
                    elif rule == 'gt':
                        query = query.filter(column > value)
                    elif rule == 'lt':
                        query = query.filter(column < value)
                    elif rule == 'gte':
                        query = query.filter(column >= value)
                    elif rule == 'lte':
                        query = query.filter(column <= value)
                    elif rule == 'contain':
                        text = f"%{value}%"
                        query = query.filter(column.like(text))
                    elif rule == 'f_contain':
                        text = f"%{value}%"
                        query = query.filter(not_(column.like(text)))
                    elif rule == 'sort':
                        if value == 'desc':
                            query = query.order_by(column.desc())
                        else:
                            query = query.order_by(column)
        return True, query

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        if self.db_model is False:
            return False, '数据库链接错误'
        flag, query = self.gen_extract_rules(self.db_model)
        if not flag:
            return False, query
        total = query.count()
        query = query.offset((page - 1) * pagesize)
        query = query.limit(pagesize)
        obj_list = query.all()
        data_li = []
        for obj in obj_list:
            dic = {c.name: getattr(obj, c.name) for c in self.table.columns}
            data_li.append(dic)
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :return:
        '''
        if not self.db_model:
            yield False, '数据库链接错误'
        flag, query = self.gen_extract_rules(self.db_model)
        if not flag:
            yield False, query
        total = query.count()
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            query = query.offset((page - 1) * pagesize)
            query = query.limit(pagesize)
            obj_list = query.all()
            data_li = []
            for obj in obj_list:
                dic = {c.name: getattr(obj, c.name) for c in self.table.columns}
                data_li.append(dic)
            res_data = {
                'records': data_li,
                'total': total
            }
            yield True, gen_json_response(res_data)

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert', 'update', 'upsert']:
            return False, f'写入类型参数错误,不支持类型{self.load_type}'
        self.only_fields = self._load_info.get('only_fields', [])
        if self.table is None:
            return False, '表不存在'
        columns = self.table.columns
        records = []
        if isinstance(res_data, list) and res_data != []:
            records = res_data
        if isinstance(res_data, dict):
            if 'records' in res_data and res_data['records'] != []:
                records = res_data['records']
            else:
                records = [res_data]
        try:
            insert_records = []
            if self.load_type == 'insert':
                insert_records = records
            elif self.load_type in ['update', 'upsert']:
                for c in records:
                    query_dict = {k: v for k, v in c.items() if k in self.only_fields}
                    query = self.session.query(self.db_model)
                    for k in query_dict:
                        query = query.filter(getattr(self.db_model, k) == query_dict[k])
                    obj = query.first()
                    if obj is not None:
                        for k in c:
                            setattr(obj, k, c[k])
                        self.session.add(obj)
                        self.session.commit()
                    elif self.load_type == 'upsert':
                        insert_records.append(c)
            if insert_records != []:
                # 创建 insert 对象
                ins = self.table.insert()
                conn = self.db_engine.connect()
                tmp_list = [{k: v for k, v in c.items() if k in columns} for c in insert_records]
                conn.execute(ins, tmp_list)
        except Exception as e:
            return False, f'{str(e)[:100]}'
        return True, res_data



from abc import ABC, abstractmethod


class DataModel(ABC):
    default_batch_size = 10000

    def __init__(self, model_info):
        self.model_info = model_info
        self._source = model_info.get('source', {})
        self._model = model_info.get('model', {})
        self._extract_info = model_info.get('extract_info', {})
        self.batch_size = self._extract_info.get('batch_size', 1000)
        self.extract_rules = self._extract_info.get('extract_rules', [])
        self._load_info = model_info.get('load_info', {})

    @abstractmethod
    def connect(self):
        '''
        连接测试
        '''
        pass

    def create(self):
        '''
        创建
        '''
        pass

    def check_field(self):
        '''
        检察字段是否存在且一致
        '''
        pass

    def set_field(self):
        '''
        设置字段
        '''
        pass

    def get_res_fields(self):
        '''
        获取字段列表
        :return:
        '''
        pass

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选规则选项
        :return:
        '''
        return []

    def parse_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        pass

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :return:
        '''
        pass

    def read_batch(self):
        '''
        分批读取数据
        :return:
        '''
        pass

    def write(self, res_data):
        '''
        写入数据
        :return:
        '''
        pass

    def update(self):
        '''
        更新数据
        :return:
        '''
        pass

    def delete(self):
        '''
        删除数据
        :return:
        '''
        pass

    def delete_batch(self):
        '''
        批量删除数据
        :return:
        '''
        pass



'''
使用代码读取mysql，转换后写入elasticsearch
'''
from ezetl.utils import get_reader, get_writer

reader_info = {
    'source': {
        "name": "test",
        "type": "mysql",
        "conn_conf": {
            "host": "127.0.0.1",
            "port": 3306,
            "username": "root",
            "password": "123456",
            "database_name": "test"
        },
        "ext_params": {}
    },
    'model': {
        "name": "btc_history",
        "type": "mysql_table",
        "model_conf": {
            "name": "btc_history",
            "auth_type": "create,insert"
        },
        "ext_params": {},
        "fields": []
    },
    'extract_info': {
        'batch_size': 100,
        'extract_rules': []
    }
}
flag, reader = get_reader(reader_info)
print(flag, reader)
flag, res = reader.connect()
print(flag, res)
print(reader.get_res_fields())
read_data = reader.read_page()
print(read_data)
load_info = {
    'source': {
        "name": "test",
        "type": "elasticsearch",
        "conn_conf": {
            "url": "127.0.0.1:9200",
            "auth_type": 2,
            "username": "elastic",
            "password": "elastic"
        },
        "ext_params": {}
    },
    'model': {
        "name": "btc_history",
        "type": "elasticsearch_index",
        "model_conf": {
            "name": "btc_history",
            "auth_type": "create,insert"
        }
    },
    'load_info': {
        'load_type': 'insert',
        'only_fields': []
    }
}
_, writer = get_writer(load_info)
print(writer)
flag, res = writer.connect()
print(flag, res)
if not flag:
    # 不存在，新建
    field_arr = [
        {'field_name': 'close', 'field_value': 'close', 'type': 'float'},
        {'field_name': 'high', 'field_value': 'high', 'type': 'float'},
        {'field_name': 'id', 'field_value': 'id', 'type': 'keyword'},
        {'field_name': 'low', 'field_value': 'low', 'type': 'float'},
        {'field_name': 'open', 'field_value': 'open', 'type': 'float'},
        {'field_name': 'symbol', 'field_value': 'symbol', 'type': 'text'},
        {'field_name': 'time', 'field_value': 'time', 'type': 'text'},
        {'field_name': 'timeframe', 'field_value': 'timeframe', 'type': 'keyword'},
        {'field_name': 'volume', 'field_value': 'volume', 'type': 'float'}
    ]
    flg, res = writer.create(field_arr)
    print(flg, res)
print(writer.get_res_fields())
for flag, read_data in reader.read_batch():
    print(read_data)
    if read_data['code'] == 200:
        write_data = read_data['data']['records']
        write_data = [i for i in write_data]
        flg, res = writer.write(write_data)
        print(flg, res)


import json
import mongoengine
from ezetl.data_models import DataModel
from ezetl.utils.common_utils import trans_rule_value, gen_json_response, parse_json


class MongoModel(DataModel):

    def __init__(self, model_info):
        super().__init__(model_info)
        conn_conf = self._source['conn_conf']
        model_conf = self._model.get('model_conf', {})
        conn_url = f"mongodb://{conn_conf.get('username')}:{conn_conf.get('password')}@{conn_conf.get('host')}:{conn_conf.get('port')}/{conn_conf.get('database_name')}"
        self.conn = mongoengine.connect(host=conn_url)
        self.collection = model_conf.get('name', '')
        if self.collection != '':
            class Model(mongoengine.DynamicDocument):
                meta = {'collection': model_conf.get('name', '')}
            self.model = Model
        else:
            self.model = None

    def connect(self):
        '''
        连通性测试
        '''
        try:
            db = mongoengine.connection.get_db()
            collection_names = db.list_collection_names()
            if self.collection == '':
                if db:
                    return True, '连接成功'
                else:
                    return False, '连接失败'
            else:
                if self.collection in collection_names:
                    return True, '连接成功'
                else:
                    return False, '连接失败'
        except Exception as e:
            print(e)
            return False, str(e)[:100]

    def get_res_fields(self):
        '''
        获取字段列表
        '''
        res_fields = []
        try:
            pass
        except Exception as e:
            print(e)
        return res_fields

    def create(self, field_arr=[]):
        '''
        创建表
        '''
        try:
            return True, '创建成功'
        except Exception as e:
            return False, str(e)

    def delete(self):
        '''
        删除
        '''
        try:
            db = mongoengine.connection.get_db()
            table = getattr(db, self.collection)
            table.drop()
            return True, '删除成功'
        except Exception as e:
            return False, str(e)

    def set_field(self, field, options={}):
        '''
        设置字段
        '''
        try:
            return True, '操作成功'
        except Exception as e:
            return False, str(e)

    def get_search_type_list(self):
        '''
        获取可用高级查询类型
        '''
        return []

    def get_extract_rules(self):
        '''
        获取可筛选项
        :return:
        '''
        rules = [{
            'name': '等于',
            'value': 'equal'
          }, {
            'name': '不等于',
            'value': 'f_equal'
          }, {
            'name': '包含',
            'value': 'contain'
          }, {
            'name': '不包含',
            'value': 'f_contain'
          }, {
            'name': '大于',
            'value': 'gt'
          }, {
            'name': '大于等于',
            'value': 'gte'
          }, {
            'name': '小于',
            'value': 'lt'
          }, {
            'name': '小于等于',
            'value': 'lte'
          }, {
            'name': '从大到小排序',
            'value': 'sort_desc'
          }, {
            'name': '从小到大排序',
            'value': 'sort_asc'
          }
        ]
        return rules

    def gen_extract_rules(self):
        '''
        解析筛选规则
        :return:
        '''
        if self.model is None:
            return False, '表不存在'
        query = self.model.objects()
        for i in self.extract_rules:
            field = i.get('field')
            rule = i.get('rule')
            value = i.get('value')
            value = trans_rule_value(value)
            if field and field not in ['sql', 'search_key']:
                if value:
                    q_dict = {}
                    if rule == 'equal':
                        q_dict = {field: value}
                    elif rule == 'f_equal':
                        q_dict = {f"{field}__ne": value}
                    elif rule == 'gt':
                        q_dict = {f"{field}__gt": value}
                    elif rule == 'lt':
                        q_dict = {f"{field}__lt": value}
                    elif rule == 'gte':
                        q_dict = {f"{field}__gte": value}
                    elif rule == 'lte':
                        q_dict = {f"{field}__lte": value}
                    elif rule == 'contain':
                        q_dict = {f"{field}__in": value}
                    elif rule == 'f_contain':
                        q_dict = {f"{field}__nin": value}
                    if q_dict != {}:
                        query = query.filter(**q_dict)
                else:
                    if rule == 'sort_asc':
                        query = query.order_by(field)
                    elif rule == 'sort_desc':
                        query = query.order_by(f"-{field}")
        return True, query

    def read_page(self, page=1, pagesize=20):
        '''
        分页读取数据
        :param page:
        :param pagesize:
        :return:
        '''
        if self.model is False:
            return False, '数据库链接错误'
        flag, query = self.gen_extract_rules()
        if not flag:
            return False, query
        total = query.count()
        query = query.skip((page - 1) * pagesize)
        query = query.limit(pagesize)
        obj_list = query.all()
        self.conn.close()
        data_li = []
        for obj in obj_list:
            dic = json.loads(obj.to_json())
            data_li.append(dic)
        res_data = {
            'records': data_li,
            'total': total
        }
        return True, gen_json_response(res_data)

    def read_batch(self):
        '''
        生成器分批读取数据
        :param res_type: 返回形式
        :return:
        '''
        if self.model is False:
            yield False, '数据库链接错误'
        flag, query = self.gen_extract_rules()
        if not flag:
            yield False, query
        total = query.count()
        pagesize = self._extract_info.get('batch_size', 1000)
        total_pages = total // pagesize + 1
        n = 0
        while n < total_pages:
            page = n + 1
            n += 1
            query = query.skip((page - 1) * pagesize)
            query = query.limit(pagesize)
            obj_list = query.all()
            data_li = []
            for obj in obj_list:
                dic = json.loads(obj.to_json())
                data_li.append(dic)
            res_data = {
                'records': data_li,
                'total': total
            }
            yield True, gen_json_response(res_data)
        self.conn.close()

    def write(self, res_data):
        self.load_type = self._load_info.get('load_type', '')
        if self.load_type not in ['insert', 'update', 'upsert']:
            return False, '写入类型参数错误'
        self.only_fields = self._load_info.get('only_fields', [])
        if self.model is None:
            return False, '表不存在'
        contents = []
        if isinstance(res_data, list) and res_data != []:
            contents = res_data
        if isinstance(res_data, dict):
            if 'contents' in res_data and res_data['contents'] != []:
                contents = res_data['contents']
            else:
                contents = [res_data]
        try:
            insert_contents = []
            if self.load_type == 'insert':
                insert_contents = contents
            elif self.load_type in ['update', 'upsert']:
                for c in contents:
                    query_dict = {k: v for k, v in c.items() if k in self.only_fields}
                    query = self.model.objects()
                    for k in query_dict:
                        query = query.filter(**{k: query_dict[k]})
                    obj = query.first()
                    if obj is not None:
                        for k in c:
                            setattr(obj, k, c[k])
                        obj.save()
                    elif self.load_type == 'upsert':
                        insert_contents.append(c)
            if insert_contents != []:
                # 创建 insert 对象
                for c in insert_contents:
                    obj = self.model(**c)
                    obj.save()
        except Exception as e:
            return False, f'{str(e)[:100]}'
        return True, res_data

